# ************************************************************************
# * Copyright (c) infrae in 2002
# * For license details please see the license.txt file. 
# *
# * Project: DOC to Silva XML Converter
# *
# * Creation Date: June 2002
# *
# * Authors: Stephan Richter, Christian Zagrodnick
# ************************************************************************
"""

$Id: P.py,v 1.23 2003/06/20 17:22:04 guido Exp $
"""

import re
from copy import copy

from util import replaceMap, stripControlChars
from silva2word.WordConstants import *
from word2silva.SilvaObject import SilvaObject

p_INDEX = re.compile(r' "([^"]+)" ')
p_SoftBreak = re.compile('\x0b+')


class Tag:
    
    def __init__(self, name, attributes, on, position):
        self.name = name
        self.attributes = attributes
        self.on = on
        self.position = position

    def tag(self):
        space = ''
        empty = ''
        if self.on in (True, None):
            end = ''
            attr = self.attributes
            if attr:
                space = ' '
            if self.on is None:
                empty = '/'
        elif self.on is False:
            end = '/'
            attr = ''
        return '<%s%s%s%s%s>' % (end, self.name, space, attr, empty)

    def __eq__(self, other):
        return self.name == other.name and self.attributes == other.attributes

    def __repr__(self):
        return "<`%s' at %d>" % (self.tag(), self.position)

class P(SilvaObject):
    """This implements the Markup type as specified in the schema"""

    debugfp = open(r'c:\P_debug', 'a')

    # map Word styles to silva styles
    style_map = {
        'Bold': 'strong',
        'Italic': 'em',
        'Underline': 'underline',
        'Superscript': 'super',
        'Subscript': 'sub',
    }
    

    __handled_fields = [wdFieldIndexEntry, wdFieldHyperlink]

    def __init__(self, original, handle_styles=1, correct_first=0):
        #self.debugfp.write('__init__\n')

        SilvaObject.__init__(self, original)

        self.text = u''
        self.markup = []
        self._non_indexes = [] 
        self._non_indexes_sorted = True

        self.handleWordParagraph(original, handle_styles, 0)

    def handleWordParagraph(self, wp, handle_styles=1, correct_first=0):
        #self.debugfp.write('\nGoing to handle %s\n' % str(wp).encode('ascii', 'replace'))
        #self.debugfp.write('Handle styles: %s\n' % (not not handle_styles))
       
        def merge(list1, list2):
            """merge two insertions lists, preserving sorting""" 
            sort_dict = {}
            for tag in list1+list2:
                sort_dict.setdefault(tag.position, []).append(tag)
            indexes = sort_dict.keys()
            indexes.sort()
            result = []
            for index in indexes:
                for tag in sort_dict[index]:
                    result.append(tag)
            return result                    
        
        if wp is None: 
            return
        
        text = wp.Range()
        wp_start = wp.Range.Start
        wp_end = wp.Range.End

        #self.debugfp.write('Text from paragraph: %s\n\n' % text)
        #self.debugfp.write('Text start: %s\n' % wp_start)
        #self.debugfp.write('Text end: %s\n' % wp_end)
        
        if stripControlChars(text.strip()) == '':
            return
        xml = u''
        inserts = []

        if correct_first:
            # somehow the first paragraph of each document needs 1 char more 
            # room than the rest
            self.nonIndex(wp_start, wp_start + 1)

        if handle_styles:
            #self.debugfp.write('Going to handle styles\n')
            inserts = self.processStyles(wp)
            #self.debugfp.write('Going to handle indexes\n')
            inserts = merge(inserts, self.processIndexes(wp))
            #self.debugfp.write('Going to handle links\n')
            inserts = merge(inserts, self.processLinks(wp))
        inserts = self.wellform(inserts)
        #self.debugfp.write('Inserts: %s\n' % inserts)
        self._blank_indexes(wp)
        self.setupKitchen()
        for tag in inserts:
            if tag.position < wp_start:
                #self.debugfp.write('Start position corrected\n')
                tag.position = wp_start
            elif tag.position > wp_end:
                #self.debugfp.write('End position corrected\n')
                tag.position = wp_end
            tag.position -= wp_start
            tp = tag.position
            tag.position = self.cookIndex(tag.position)
            #self.debugfp.write('Before cookIndex: %s, after: %s\n' % (tp, tag.position))
            #if correct_first:
            #    #self.debugfp.write('Correcting first item\n')
            #    tag.position -= 1

        next = 0
        for tag in inserts:
            #self.debugfp.write('Tag: %s, pos: %s\n' % (tag, tag.position))
            xml += replaceMap(text[next:tag.position])
            xml += tag.tag()
            next = tag.position
            #self.debugfp.write('XML so far: %s\n' % xml)

        xml += replaceMap(text[next:])
        xml = p_SoftBreak.sub('<br/>', xml)
        xml = stripControlChars(xml)
                
        self.text = xml

    def processComments(self, wp):
        """removes comments"""

        #self.debugfp.write('processStyles\n')

        wp_start = wp.Range.Start
        wp_end = wp.Range.End

        #self.debugfp.write('Start wp: %s\n' % wp_start)
        #self.debugfp.write('End wp: %s\n' % wp_end)

        for field in wp.Range.Fields:
            if not field.type == wdFieldComments:
                continue

            start = field.Code.start
            end = field.Code.end

            self.nonIndex(start, end)

            #self.debugfp.write('Comment removed from %s, %s\n' % start, end)


    def processStyles(self, wp):
        """processes markup like bold and italic"""

        #self.debugfp.write('processStyles\n')

        styles = []
        styles_dict = {}
        wp_start = wp.Range.Start
        wp_end = wp.Range.End

        #self.debugfp.write('Start wp: %s\n' % wp_start)
        #self.debugfp.write('End wp: %s\n' % wp_end)

        for word_style, silva_style in self.style_map.items():
            wp_range = wp.Range
            find = wp_range.Find
            self.clearFind(find)
            setattr(find.Font, word_style, True)
            last_range = None
            while find.Execute(Replace=wdReplaceNone):
                if (wp_range.Start, wp_range.End) == last_range:
                    break
                last_range = (wp_range.Start, wp_range.End)
                #self.debugfp.write('Range start: %s, end: %s\n' % (wp_range.Start, wp_range.End))
                if wp_range.Start < wp_start:
                    #self.debugfp.write('Range start < wp_start\n')
                    continue
                if wp_range.End >= wp_end:
                    #self.debugfp.write('Range end > wp_end\n')
                    break
                styles_dict.setdefault(wp_range.Start, []).append(
                    (self.style_map[word_style], True))
                styles_dict.setdefault(wp_range.End, []).insert(0, 
                    (self.style_map[word_style], False))
                #self.debugfp.write('Tag placed on %s, %s\n' % (wp_range.Start, wp_range.End))
        index_list = styles_dict.keys()
        index_list.sort()
        for index in index_list:
            styles += map(lambda (style, onoff): Tag(style, '', onoff, index),
                 styles_dict[index])
        return styles

    def processIndexes(self, wp):
        #debugfp = open(r'c:\index_debug', 'a')
        #self.debugfp.write('processIndexes\n')
            
        if len(wp.Range.Fields) == 0:
            #debugfp.write('No fields\n')
            return []
    
        wp_start = wp.Range.Start
        #debugfp.write('Wp start: %s\n' % wp_start)
        result = []
        for field in wp.Range.Fields:
            if not field.type == wdFieldIndexEntry: 
                #debugfp.write('No fieldindexentry\n')
                continue 
            
            start = field.Code.Start
            end = field.Code.End

            #debugfp.write('Start of code: %s\n' % start)
            #debugfp.write('End of code: %s\n' % end)
            #debugfp.write('End - 2: %s\n' % wp.Range.Document.Range(end - 2, end - 2)())
            #debugfp.write('Field content: %s\n' % field.Code())
            #debugfp.write('Field content: %s\n' % wp.Range.Document.Range(start, end))
            
            # somehow this needs 1 character extra space
            self.nonIndex(start, end + 1)
            
            m = p_INDEX.search(field.Code())
            if m is None: 
                #debugfp.write('No index\n')
                continue
            name = m.group(1)
            #debugfp.write('Added name %s, loc %s\n' % (name, end))
            
            tag = Tag('index', 'name="%s"' % (replaceMap(name), ), None, end)
            result.append(tag)
   
        return result

    def processLinks(self, wp):

        if len(wp.Range.Hyperlinks) == 0:
            return []
        
        result = []
        for link in wp.Range.Hyperlinks:
            
            start = link.Range.start
            end = link.Range.end
            
            #self.debugfp.write('%s\n' % str(link.Range()))

            #self.debugfp.write('Start: %s\n' % start)
            #self.debugfp.write('End: %s\n' % end)

            #self.debugfp.write('Sel start: %s\n' % wp.Range.Start)
            #self.debugfp.write('Sel end: %s\n' % wp.Range.End)
            
            url = replaceMap(link.address)
            text = link.TextToDisplay
            
            #self.debugfp.write('Text: %s\n' % text)

            text_index = end - len(text) - 2 
            self.nonIndex(start, text_index)
            self.nonIndex(end + 1, end + 1)

            #self.debugfp.write('Text index: %s\n' % text_index)
            #self.debugfp.write('Text length: %s\n' % len(text))
            
            start_tag = Tag('link', 'url="%s"' % (replaceMap(url), ), True, text_index)
            end_tag = copy(start_tag)
            end_tag.on = False
            end_tag.position = end
            result.append(start_tag)
            result.append(end_tag)
            #self.debugfp.write('Start tag: %s\n' % start_tag)
            #self.debugfp.write('End tag: %s\n' % end_tag)
            #self.debugfp.write('Character at position of end_tag: %s\n' % (wp.Range.Document.Range(text_index, text_index + 3)))
            #self.debugfp.write('Text between start of paragraph and start of link text: \'%s\'\n' % (wp.Range.Document.Range(start, text_index)))
        return result

    def nonIndex(self, start, end):

        self._non_indexes += range(start - 1, end)
        self._non_indexes_sorted = False
   
    def setupKitchen(self):
        """sets up the kitchen for cooking indexes"""
        
        wp_start = self.original.Range.Start
        wp_end = self.original.Range.End
        if not self._non_indexes_sorted:
            self._non_indexes = dict(map(lambda x: (x, 0), 
                self._non_indexes)).keys()
            self._non_indexes.sort()
            self._non_indexes = \
                map(lambda x: x - wp_start, 
                filter(lambda x: wp_start <= x <= wp_end,
                    self._non_indexes))
            self._non_indexes_sorted = True
   
    def cookIndex(self, raw_index):
        #self.debugfp.write('Going to cook index for %s\n' % raw_index)
        nis = self._non_indexes
        #self.debugfp.write('Nis: %s\n' % nis)
        
        while raw_index in nis:
            raw_index += 1
        corrector = 0
        for i in range(0, len(nis)):
            if raw_index <= nis[i]:
                break
            else:
                corrector += 1
        #self.debugfp.write('Corrected index: %s\n' % (raw_index - corrector))
        return raw_index - corrector

    def _blank_indexes(self, wp):
        if len(wp.Range.Fields) == 0: 
            return
        for field in wp.Range.Fields:
            if not field.Type in self.__handled_fields:
                self.nonIndex(field.Code.Start, field.Code.End)

    def clearFind(self, find):
        """clear the given find instance"""
        find.ClearFormatting()
        find.ClearAllFuzzyOptions()
        find.Format = 1
        find.Text = u''
        find.MatchCase = 0
        find.MatchWholeWord = 0
        find.MatchWildcards = 0 
        find.MatchSoundsLike = 0
        find.MatchAllWordForms = 0
        find.Wrap = wdWrapNever 
        find.Forward = True

    def wellform(tags):
        """wellform the styles
        
            It basicaly handles cases which would lead to:

                <strong>Foo<em>Bar</strong>baz</em>

            By replacing it by:

                <strong>Foo<em>Bar</em></strong><em>baz</em>
        
            It's not done with xml tags but an internal structure though.

            returns a list of Tag instances
        """
        
        result = []
        stack = []
        for tag in tags:
            if tag.on is True:
                if not tag in stack:
                    result.append(tag)
                    stack.append(copy(tag))
            elif tag.on is False:
                if tag in stack:
                    deactivated_tags = []
                    while 1:
                        deactivate_tag = stack.pop()
                        deactivate_tag.on = False
                        deactivate_tag.position = tag.position
                        result.append(deactivate_tag)
                        if deactivate_tag == tag:
                            break
                        deactivated_tags.append(copy(deactivate_tag))
                    deactivated_tags.reverse()
                    for reactivate_tag in deactivated_tags:
                        reactivate_tag.on = True
                        reactivate_tag.position = tag.position
                        stack.append(copy(reactivate_tag))
                        result.append(reactivate_tag)
            elif tag.on is None:
                result.append(tag)
            else:
                raise ValueError, 'tag.on is invalid (%r)' % (tag.on,)
        assert stack == [], "Stack is not empty: %r (%r)" % (stack, result)
        return result
    wellform = staticmethod(wellform)

    def getXML(self):
        # XXX can there be more types of P here??
        if self.text:
            return u'<p type="normal">%s</p>' % self.text
        return u''

class Pre(SilvaObject):

    def getXML(self):

        elements = \
            map(stripControlChars, 
            map(lambda x: replaceMap(x),
            self._elements))
        text = '\n'.join(elements)    
            
        xml = u'<pre>%s</pre>' % (text, )
        return xml

    
